import typing as tp
import logging
import threading
from .decorators import wraps

__all__ = [
    'rethrow_as',
    'silence_excs',
    'catch_exception',
    'log_exceptions',
]

ExcType = tp.Type[Exception]
T = tp.TypeVar('T')


def silence_excs(*exc_types: ExcType, returns=None):
    """
    Silence given exception types.

    Can be either a decorator or a context manager.

    If you are using it as a decorator, you can specify what value should the function return
    by using the returns kwarg:

    >>> @silence_excs(KeyError, returns=5)
    >>> def returns_5():
    >>>     raise KeyError()
    >>> assert returns_5() == 5
    """
    return rethrow_as(exc_types, None, returns=returns)


class log_exceptions:
    """
    Decorator/context manager to log your exceptions into the log.

    The exception will be logged and re-raised.

    Logger will be passed the exception instance as exc_info.

    :param logger: a logger to which the exception has to be logged
    :param severity: a severity level
    :param format_string: a format string with fields:
        - e      : the exception instance itself
        - args   : arguments with which the function was called, unavailable if context manager
        - kwargs : arguments with which the function was called, unavailable if context manager
        You can specify additional fields providing the locals_ argument
        Example: "{exc_type} occurred with message {exc_val} with traceback {exc_tb}"
    :param locals_: local variables to add to the format string. args and kwargs will be overwritten
        by this, but e will never be overwritten.
    :param exc_types: logger will log only on those exceptions. Default is None which means
        log on all exceptions
    """
    __slots__ = ('logger', 'severity', 'format_string', 'locals', 'exc_types')

    def __init__(self, logger, severity=logging.ERROR, format_string='{exc_val}',
                 locals_: tp.Optional[tp.Dict] = None,
                 exc_types: tp.Optional[tp.Union[ExcType, tp.Sequence[ExcType]]] = None):
        self.logger = logger
        self.severity = severity
        self.format_string = format_string
        self.locals = locals_
        if exc_types is None:
            self.exc_types = Exception
        else:
            self.exc_types = exc_types

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if issubclass(exc_type, self.exc_types):
                format_dict = {}
                if self.locals is not None:
                    format_dict.update(self.locals)
                format_dict['e'] = exc_val
                self.logger.log(self.severity, self.format_string.format(format_dict),
                                exc_info=exc_val)
        return False

    def __call__(self, fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            try:
                return fun(*args, **kwargs)
            except Exception as e:
                if isinstance(e, self.exc_types):
                    format_dict = {'args': args,
                                   'kwargs': kwargs}
                    if self.locals is not None:
                        format_dict.update(self.locals)
                    format_dict['e'] = e
                    self.logger.log(self.severity, self.format_string.format(format_dict),
                                    exc_info=e)
                raise e
        return inner


class rethrow_as:
    """
    Transform some exceptions into others.

    Either a decorator or a context manager

    New exception will be created by calling exception to transform to with
    repr of current one.

    You can also provide just two exceptions, eg.

    >>> rethrow_as(NameError, ValueError)

    You can also provide a pairwise translation, eg. from NameError to ValueError and from OSError
    to IOError

    >>> rethrow_as((NameError, ValueError), (OSError, IOError))

    If the second value is a None, exception will be silenced.
    """
    __slots__ = ('mapping', 'exception_preprocessor', 'returns', '__exception_remapped')

    def __init__(self, *pairs: tp.Union[ExcType, tp.Tuple[ExcType, ExcType]],
                 exception_preprocessor: tp.Optional[tp.Callable[[Exception], str]] = repr,
                 returns=None):
        """
        Pass tuples of (exception to catch - exception to transform to).

        :param exception_preprocessor: other callable/1 to use instead of repr.
            Should return a str, a text description of the exception
        :param returns: what value should the function return if this is used as a decorator
        """
        try:
            a, b = pairs  # throws ValueError
            op = issubclass(b, BaseException)  # throws TypeError
        except TypeError:
            op = b is None
        except ValueError:
            op = False

        if op:
            pairs = [pairs]

        self.mapping = list(pairs)
        self.exception_preprocessor = exception_preprocessor or repr
        self.returns = returns

        # this is threading.local because two threads may execute the same function at the
        # same time, and exceptions from one function would leak to another
        self.__exception_remapped = threading.local()

    def __call__(self, fun: tp.Callable) -> tp.Callable:
        @wraps(fun)
        def inner(*args, **kwargs):
            with self:
                v = fun(*args, **kwargs)
            if self.__exception_remapped.was_raised:
                # This means that the normal flow of execution was interrupted
                return self.returns
            else:
                return v

        return inner

    def __enter__(self):
        self.__exception_remapped.was_raised = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            for from_, to in self.mapping:
                if issubclass(exc_type, from_):
                    self.__exception_remapped.was_raised = True
                    if to is None:
                        return True
                    else:
                        raise to(self.exception_preprocessor(exc_val))


def catch_exception(exc_class: tp.Union[ExcType, tp.Tuple[ExcType]],
                    clb: tp.Callable[[], T],
                    return_instead: tp.Optional[T] = None,
                    return_value_on_no_exception: bool = False) -> tp.Union[Exception, T]:
    """
    Catch exception of given type and return it. Functionally equivalent to:

    >>> try:
    >>>     v = clb()
    >>>     if return_value_on_no_exception:
    >>>         return v
    >>> except exc_class as e:
    >>>     if return_instead:
    >>>         return return_instead
    >>>     return e

    :param exc_class: Exception classes to catch
    :param clb: callable/0 to call to raise the exception
    :param return_instead: what to return instead of the function result if it didn't end in an
        exception
    :param return_value_on_no_exception: whether to return the function result if exception didn't
        happen
    :raises ValueError: an exception was not thrown
    :raises TypeError: a different exception was thrown that the one we're catchin
    """
    try:
        result = clb()
    except exc_class as e:
        return e
    except Exception as e:
        raise TypeError('%s was thrown instead' % (e,))

    if return_instead is not None:
        return return_instead

    if return_value_on_no_exception:
        return result

    raise ValueError('Callable executed without error')



from .dictionaries import DictObject, apply_dict_object, DictionaryView, TwoWayDictionary, \
    DirtyDict, KeyAwareDefaultDict
from .hashable_objects import HashableWrapper
from .immutable import Immutable, frozendict
from .singleton import Singleton, SingletonWithRegardsTo
from .structures import SetHeap, OmniHashableMixin, TimeBasedHeap, Heap, ReprableObject
from .proxy import Proxy
from .typednamedtuple import typednamedtuple
from .ranking import Ranking
from .sorted_list import SortedList, SliceableDeque

__all__ = [
    'KeyAwareDefaultDict',
    'Proxy',
    'ReprableObject',
    'DirtyDict',
    'SortedList',
    'SliceableDeque',
    'Ranking',
    'TwoWayDictionary',
    'typednamedtuple',
    'HashableWrapper',
    'Heap',
    'SetHeap',
    'DictionaryView',
    'frozendict',
    'TimeBasedHeap',
    'OmniHashableMixin',
    'Singleton',
    'SingletonWithRegardsTo',
    'DictObject',
    'apply_dict_object',
    'Immutable'
]



import json
import typing as tp
from abc import ABCMeta, abstractmethod

__all__ = ['JSONEncoder', 'JSONAble', 'json_encode']


Jsonable = tp.TypeVar('Jsonable', list, dict, str, int, float, None)


class JSONAble(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def to_json(self) -> Jsonable:
        """Return a JSON-able representation of this object"""


class JSONEncoder(json.JSONEncoder):
    """
    This encoder will encode everything!
    """
    def default(self, o: tp.Any) -> Jsonable:
        if hasattr(o, 'to_json'):
            return o.to_json()
        try:
            return super().default(o)
        except TypeError:
            dct = {}
            for k, v in o.__dict__.items():
                dct[k] = repr(v)
            return dct


def json_encode(x: tp.Any) -> str:
    """
    Convert an object to JSON. Will properly handle subclasses of JSONAble

    :param x: object to convert
    """
    return JSONEncoder().encode(x)



import typing as tp
from .base import EmbeddedSubmetrics, MetricLevel
from .measurable_mixin import MeasurableMixin
from ..data import MetricData, MetricDataCollection
from .registry import register_metric


@register_metric
class CounterMetric(EmbeddedSubmetrics, MeasurableMixin):
    """
    A counter that can be adjusted by a given value.

    :param sum_children: whether to sum up all calls to children
    :param count_calls: count the amount of calls to handle()
    """
    __slots__ = ('sum_children', 'count_calls', 'calls', 'value')

    CLASS_NAME = 'counter'

    def __init__(self, name, root_metric: 'Metric' = None,
                 metric_level: tp.Optional[MetricLevel] = None,
                 internal: bool = False,
                 sum_children: bool = True,
                 count_calls: bool = False, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, internal=internal,
                         sum_children=sum_children, count_calls=count_calls, *args, **kwargs)
        self.sum_children = sum_children            # type: bool
        self.count_calls = count_calls              # type: bool
        self.calls = 0                              # type: int
        self.value = 0                              # type: float

    def to_metric_data(self) -> MetricDataCollection:
        if self.embedded_submetrics_enabled:
            k = super().to_metric_data()
            if self.sum_children:
                k += MetricData(self.name+'.sum', self.value, self.labels, self.get_timestamp(),
                                self.internal)
            if self.count_calls:
                k += MetricData(self.name+'.count', self.calls, self.labels, self.get_timestamp(),
                                self.internal)
            return k

        p = super().to_metric_data()
        p.set_value(self.value)
        if self.count_calls:
            p += MetricData(self.name+'.count', self.calls, self.labels, self.get_timestamp(),
                            self.internal)

        return p

    def _handle(self, delta: float = 0, **labels):
        if self.embedded_submetrics_enabled or labels:
            if self.sum_children:
                self.value += delta
            self.calls += 1
            return super()._handle(delta, **labels)

        self.value += delta
        self.calls += 1



import collections
import copy

import threading
import typing as tp

from ..decorators import wraps

__all__ = [
    'Monitor', 'RMonitor', 'MonitorDict', 'MonitorList'
]

K, V, T = tp.TypeVar('K'), tp.TypeVar('V'), tp.TypeVar('T')



class Monitor:
    """
    Base utility class for creating monitors (the synchronization thingies!)

    These are NOT re-entrant!

    Use it like that:

    >>> class MyProtectedObject(Monitor):
    >>>     def __init__(self, *args, **kwargs):
    >>>         Monitor.__init__(self)
    >>>         ... do your job ..

    >>>     @Monitor.synchronized
    >>>     def function_that_needs_mutual_exclusion(self):
    >>>         .. do your threadsafe jobs ..

    >>>     def function_that_partially_needs_protection(self):
    >>>         .. do your jobs ..
    >>>         with Monitor.acquire(self):
    >>>             .. do your threadsafe jobs ..
    >>>         .. do your jobs ..
    >>>         with self:
    >>>             .. do your threadsafe jobs ..
    """
    __slots__ = ('_monitor_lock',)

    def __enter__(self):
        self._monitor_lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._monitor_lock.release()
        return False

    def __init__(self):
        """You need to invoke this at your constructor
        You can also use it to release locks of other objects."""
        self._monitor_lock = threading.Lock()  # type: threading.Lock

    @staticmethod
    def synchronized(fun):
        """
        This is a decorator. Class method decorated with that will lock the
        global lock of given instance, making it threadsafe. Depending on
        usage pattern of your class and it's data semantics, your performance
        may vary
        """

        @wraps(fun)
        def monitored(*args, **kwargs):
            # noinspection PyProtectedMember
            with args[0]._monitor_lock:
                return fun(*args, **kwargs)

        return monitored

    class release:
        """
        Returns a context manager object that can release another object
        as long as that object is a monitor.

        Consider foo, which is a monitor. You have a protected function,
        but you feel that you can release it for a while as it would
        improve parallelism. You can use it as such:

        >>> @Monitor.synchronized
        >>> def protected_function(self):
        >>>     .. do some stuff that needs mutual exclusion ..
        >>>     with Monitor.release(self):
        >>>         .. do some I/O that doesn't need mutual exclusion ..
        >>>     .. back to protected stuff ..
        """

        def __init__(self, foo: 'Monitor'):
            self.foo = foo

        def __enter__(self):
            # noinspection PyProtectedMember
            self.foo._monitor_lock.release()

        def __exit__(self, e1, e2, e3):
            # noinspection PyProtectedMember
            self.foo._monitor_lock.acquire()
            return False

    class acquire:
        """
        Returns a context manager object that can lock another object,
        as long as that object is a monitor.

        Consider foo, which is a monitor. If you needed to lock it from
        outside, you would do:

        >>> with Monitor.acquire(foo):
        >>>     .. do operations on foo that need mutual exclusion ..
        """

        def __init__(self, foo: 'Monitor'):
            self.foo = foo

        def __enter__(self):
            # noinspection PyProtectedMember
            self.foo._monitor_lock.acquire()

        def __exit__(self, e1, e2, e3):
            # noinspection PyProtectedMember
            self.foo._monitor_lock.release()
            return False

    @classmethod
    def synchronize_on(cls, monitor):
        """
        A decorator for locking on non-self Monitor objects

        Use it like:

        >>> class MasterClass(Monitor):
        >>>     def get_object(self):
        >>>         class SlaveClass:
        >>>             @Monitor.synchronize_on(self)
        >>>             def get_object(self2):
        >>>                 ...
        >>>         return SlaveClass
        """

        def outer(fun):
            @wraps(fun)
            def inner(*args, **kwargs):
                with cls.acquire(monitor):
                    return fun(*args, **kwargs)

            return inner

        return outer


class RMonitor(Monitor):
    """
    Monitor, but using an reentrant lock instead of a normal one
    """

    def __init__(self):
        self._monitor_lock = threading.RLock()  # type: threading.RLock


class MonitorList(tp.Generic[T], collections.UserList, Monitor):
    """A list that is also a monitor"""

    def __init__(self, *args):
        collections.UserList.__init__(self, *args)
        Monitor.__init__(self)

    def __copy__(self):
        return MonitorList(copy.copy(self.data))

    def __deepcopy__(self, memo):
        return MonitorList(copy.deepcopy(self.data, memo))

    def __getitem__(self, item: tp.Union[slice, int]) -> T:
        return self.data[item]

    def __setitem__(self, key: int, value: T) -> None:
        self.data[key] = value

    def __delitem__(self, key: tp.Union[slice, int]) -> None:
        del self.data[key]


class MonitorDict(tp.Generic[K, V], collections.UserDict, Monitor):
    """A dict that is also a monitor"""

    def __init__(self, *args, **kwargs):
        collections.UserDict.__init__(self, *args, **kwargs)
        Monitor.__init__(self)

    def __getitem__(self, item: K) -> V:
        return self.data[item]

    def __setitem__(self, key: K, value: V) -> None:
        self.data[key] = value

    def __delitem__(self, key: K) -> None:
        del self.data[key]

    def __copy__(self):
        return MonitorDict(copy.copy(self.data))

    def __deepcopy__(self, memo):
        return MonitorDict(copy.deepcopy(self.data, memo))



from .choose import choose
from .iterators import infinite_counter, take_n, is_instance, skip_first, zip_shifted, stop_after, \
    iter_dict_of_list, shift, other_sequence_no_longer_than, count, even, odd, n_th
from .sequences import is_last, add_next, half_cartesian, group_quantity, Multirun

__all__ = ['choose', 'infinite_counter', 'take_n', 'is_instance', 'is_last', 'add_next',
           'half_cartesian', 'skip_first', 'zip_shifted', 'stop_after', 'group_quantity',
           'iter_dict_of_list', 'shift', 'other_sequence_no_longer_than', 'count', 'n_th',
           'even', 'odd', 'Multirun']



import logging
import os
import time
import typing as tp

import psutil

from satella.coding.concurrent import CallableGroup, CallNoOftenThan
from satella.coding.concurrent import TerminableThread
from satella.coding.structures import Singleton
from satella.time import measure
from .conditions import BaseCondition, ZerothSeverity

logger = logging.getLogger(__name__)

__all__ = ['MemoryPressureManager']


@Singleton
class MemoryPressureManager(TerminableThread):
    """
    Manager of the memory pressure.

    The program is in some severity state. The baseline state is 0, meaning everything's OK.

    Please note that it is sufficient to instantiate this class for the thread to run.

    Eg.

    >>> mt = MemoryPressureManager(maximum_available=4*GB, severity_levels=[80, 90])
    >>> @mt.register_on_severity(1)
    >>> def trigger_a():
    >>>     print('80% consumption of memory exceeded')
    >>> @mt.register_on_severity(2)
    >>> def trigger_b():
    >>>     print('90% consumption of memory exceeded')

    :param maximum_available: maximum amount of memory that this program can use
    :param severity_levels: this defines the levels of severity. A level is reached when program's
        consumption is other this many percent of it's maximum_available amount of memory.
    :param check_interval: amount of seconds of pause between consecutive checks
    :param log_transitions: whether to log to logger when a transition takes place
    """
    __slots__ = ('process', 'maximum_available', 'severity_levels', 'callbacks_on_entered',
                 'callbacks_on_remains', 'current_severity_level', 'check_interval',
                 'log_transitions')

    def __init__(self, maximum_available: tp.Optional[int] = None,
                 severity_levels: tp.List[BaseCondition] = None,
                 check_interval: int = 10,
                 log_transitions: bool = True):
        super().__init__(daemon=True)
        self.log_transitions = log_transitions          # type: bool
        self.process = psutil.Process(os.getpid())      # type: psutil.Process
        self.maximum_available = maximum_available      # type: int
        self.severity_levels = [ZerothSeverity()] + (
                severity_levels or [])                  # type: tp.List[BaseCondition]

        self.callbacks_on_entered = [CallableGroup(gather=False) for _ in
                                     range(len(
                                         self.severity_levels))]  # type: tp.List[CallableGroup]
        self.callbacks_on_remains = [CallableGroup(gather=False) for _ in
                                     range(len(
                                         self.severity_levels))]  # type: tp.List[CallableGroup]
        self.callbacks_on_left = [CallableGroup(gather=False) for _ in
                                  range(len(
                                      self.severity_levels))]  # type: tp.List[CallableGroup]
        self.current_severity_level = 0             # type: int
        self.stopped = False                        # type: bool
        self.check_interval = check_interval        # type: int
        self.start()

    def advance_to_severity_level(self, target_level: int):
        while self.current_severity_level != target_level:
            delta = target_level - self.current_severity_level
            delta = int(delta / abs(delta))

            if delta > 0:
                # Means we are ENTERING a severity level
                self.current_severity_level += delta
                self.callbacks_on_entered[self.current_severity_level]()
                if self.log_transitions:
                    logger.warning('Entered severity level %s' % (self.current_severity_level, ))
            elif delta < 0:
                # Means we are LEAVING a severity level
                self.callbacks_on_left[self.current_severity_level]()
                if self.log_transitions:
                    logger.warning('Left severity level %s' % (self.current_severity_level, ))
                self.current_severity_level += delta

    def stop(self):
        """Stop this thread from operating"""
        self.stopped = True

    def resume(self):
        """Resume the operation of this thread"""
        self.stopped = False

    def loop(self) -> None:
        if self.stopped:
            return time.sleep(self.check_interval)

        with measure() as measurement:
            severity_level = self.calculate_severity_level()
            if self.current_severity_level != severity_level:
                self.advance_to_severity_level(severity_level)
            else:
                self.callbacks_on_remains[severity_level]()

        elapsed = measurement()
        if elapsed < self.check_interval:
            time.sleep(self.check_interval - elapsed)

    def calculate_severity_level(self) -> int:
        """
        This returns a severity level. 0 is the baseline severity level.
        """
        memory_info = self.process.memory_info()
        for level, condition in reversed(list(enumerate(self.severity_levels))):
            if condition.can_fire(memory_info, self.maximum_available):
                return level

    @staticmethod
    def register_on_entered_severity(severity: int):
        """
        Register this handler to fire on entered a particular severity level.

        This means that situation has gotten worse.

        Use like this:

        >>> MemoryPressureManager.register_on_entered_severity(1)
        >>> def entered_severity_one():
        >>>     print('Entered memory severity level 1')

        :param severity: severity level to react to
        """

        def outer(fun):
            MemoryPressureManager().callbacks_on_entered[severity].add(fun)
            return fun

        return outer

    @staticmethod
    def register_on_left_severity(severity: int):
        """
        Register a handler to be called when given severity level is left. This means
        that we have advanced to a lower severity level.

        >>> MemoryPressureManager.register_on_left_severity(1)
        >>> def entered_severity_one():
        >>>     print('Memory comsumption no longer 1')


        :param severity: severity level to leave
        """
        def outer(fun):
            MemoryPressureManager().callbacks_on_left[severity].add(fun)
            return fun
        return outer

    @staticmethod
    def register_on_remaining_in_severity(severity: int, call_no_more_often_than: int = 0):
        """
        Register this handler to fire on remaining in a particular severity level. Use like this:

        >>> MemoryPressureManager.register_on_remaining_in_severity(0, 30)
        >>> def entered_severity_one():
        >>>     print('Memory comsumption OK. I am called no more often than each 30 seconds')

        :param severity: severity level
        :param call_no_more_often_than: call no more often than this amount of seconds
        """

        def outer(fun):
            MemoryPressureManager().callbacks_on_remains[severity].add(
                CallNoOftenThan(call_no_more_often_than, fun))
            return fun

        return outer



import typing as tp

__all__ = ['is_last', 'add_next', 'half_cartesian', 'group_quantity']

from satella.coding.decorators import wraps
from .iterators import n_th

T = tp.TypeVar('T')
U = tp.TypeVar('U')


# shamelessly copied from https://medium.com/better-programming/is-this-the-last-element-of-my-python-for-loop-784f5ff90bb5
def is_last(lst: tp.Iterable[T]) -> tp.Generator[tp.Tuple[bool, T], None, None]:
    """
    Return every element of the list, alongside a flag telling is this the last element.

    Use like:

    >>> for is_last, element in is_last(my_list):
    >>> if is_last:
    >>>     ...

    :param lst: list to iterate thru
    :return: a generator returning (bool, T)
    """
    iterable = iter(lst)
    ret_var = next(iterable)
    for val in iterable:
        yield False, ret_var
        ret_var = val
    yield True, ret_var


def add_next(lst: tp.Iterable[T], wrap_over: bool = False, skip_last: bool = False) -> tp.Iterator[
        tp.Tuple[T, tp.Optional[T]]]:
    """
    Yields a 2-tuple of given iterable, presenting the next element as second element of the tuple.

    The last element will be the last element alongside with a None, if wrap_over is False, or the
    first element if wrap_over was True

    Example:

    >>> list(add_next([1, 2, 3, 4, 5])) == [(1, 2), (2, 3), (3, 4), (4, 5), (5, None)]
    >>> list(add_next([1, 2, 3, 4, 5], True)) == [(1, 2), (2, 3), (3, 4), (4, 5), (5, 1)]

    :param lst: iterable to iterate over
    :param wrap_over: whether to attach the first element to the pair of the last element instead
        of None
    :param skip_last: if this is True, then last element, alongside with a None, won't be output
    """
    iterator = iter(lst)
    try:
        first_val = prev_val = next(iterator)
    except StopIteration:
        return
    for val in iterator:
        yield prev_val, val
        prev_val = val
    if wrap_over:
        yield prev_val, first_val
    else:
        if not skip_last:
            yield prev_val, None


def half_cartesian(seq1: tp.Iterable[T], include_same_pairs: bool = True) -> tp.Iterator[tp.Tuple[T, T]]:
    """
    Generate half of the Cartesian product of both sequences.

    Useful when you have a commutative operation that you'd like to execute on both elements
    (eg. checking for collisions).

    Example:

    >>> list(half_cartesian([1, 2, 3], [1, 2, 3])) == \
    >>>     [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)]

    :param seq1: First sequence
    :param seq2: Second sequence
    :param include_same_pairs: if True, then pairs returning two of the same objects will be returned.
        Ie. if False, the following will be true:

    >>> list(half_cartesian([1, 2, 3], [1, 2, 3], include_same_pairs=False)) == \
    >>>     [(1, 2), (1, 3), (2, 3)]

    """
    for i, elem1 in enumerate(seq1):
        for j, elem2 in enumerate(seq1):
            if include_same_pairs:
                if j >= i:
                    yield elem1, elem2
            else:
                if j > i:
                    yield elem1, elem2


def group_quantity(length: int, seq: tp.Iterable[T]) -> tp.Generator[tp.List[T], None, None]:
    """
    Slice an iterable into lists containing at most len entries.

    Eg.

    >>> assert list(group_quantity(3, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])) == [[1, 2, 3], [4, 5, 6],
    >>>                                                                     [7, 8, 9], [10]]

    :param length: length for the returning sequences
    :param seq: sequence to split
    """
    entries = []
    for elem in seq:
        if len(entries) < length:
            entries.append(elem)
        else:
            yield entries
            entries = [elem]
    if entries:
        yield entries


class Multirun:
    """
    A class to launch the same operation on the entire sequence.

    Consider:

    >>> class Counter:
    >>>     def __init__(self, value=0):
    >>>         self.count = value
    >>>     def add(self, v):
    >>>         self.count += 1
    >>>     def __eq__(self, other):
    >>>          return self.count == other.count
    >>>     def __iadd__(self, other):
    >>>         self.add(other)
    >>> a = [Counter(), Counter()]

    The following:

    >>> for b in a:
    >>>     b.add(2)

    Can be replaced with

    >>> Multirun(a).add(2)

    And the following:

    >>> for b in a:
    >>>     b += 3

    With this

    >>> b = Mulirun(a)
    >>> b += 3

    Furthermore note that:

    >>> Multirun(a).add(2) == [Counter(2), Counter(2)]

    :param sequence: sequence to execute these operations for
    :param dont_return_list: the operation won't return a list if this is True
    """
    def __init__(self, sequence: tp.Iterable, dont_return_list: bool = False):
        self.sequence = sequence
        self.dont_return_list = dont_return_list

    def __iter__(self):
        return iter(self.sequence)

    def __getattr__(self, item):
        def inner(*args, **kwargs):
            if not self.dont_return_list:
                results = []
            for element in self:
                getattr(element, item)(*args, **kwargs)
                if not self.dont_return_list:
                    results.append(element)
            if not self.dont_return_list:
                return results

        # Take care: the array might just be empty...
        try:
            fun = getattr(n_th(self), item)
            inner = wraps(fun)(inner)
        except IndexError:
            pass

        return inner

    def __iadd__(self, other):
        for element in self:
            element += other
        return self

    def __isub__(self, other):
        for element in self:
            element -= other
        return self

    def __imul__(self, other):
        for element in self:
            element *= other
        return self

    def __itruediv__(self, other):
        for element in self:
            element /= other
        return self

    def __ifloordiv__(self, other):
        for element in self:
            element //= other
        return self

    def __ilshift__(self, other):
        for element in self:
            element <<= other
        return self

    def __irshift__(self, other):
        for element in self:
            element >>= other
        return self



import importlib
import os
import pkgutil
import warnings
import typing as tp

__all__ = ['import_from', 'import_class']


def import_class(path: str) -> type:
    """
    Import a class identified with given module path and class name

    :param path: path, eg. subprocess.Popen
    :return: imported class
    """
    *path, classname = path.split('.')
    import_path = '.'.join(path)
    try:
        return getattr(importlib.import_module(import_path), classname)
    except AttributeError:
        raise ImportError('%s not found in %s' % (classname, import_path))


def import_from(path: tp.List[str], package_prefix: str, all_: tp.List[str],
                locals: tp.Dict[str, tp.Any], recursive: bool = True,
                fail_on_attributerror: bool = True, create_all: bool = True,
                skip_single_underscores: bool = True,
                skip_not_having_all: bool = False) -> None:
    """
    Import everything from a given module. Append these module's all to.

    This will examine __all__ of given module (if it has any, else it will just import everything
    from it, which is probably a bad practice and will heavily pollute the namespace.

    As a side effect, this will equip all of your packages with __all__.

    :param path: module's __path__
    :param package_prefix: package prefix to import from. Use __name__
    :param all_: module's __all__ to append to
    :param recursive: whether to import packages as well
    :param fail_on_attributerror: whether to fail if a module reports something in their __all__ that
        is physically not there (ie. getattr() raised AttributeError
    :param locals: module's locals, obtain them by calling locals() in importing module's context
    :param create_all: whether to create artificial __all__'s for modules that don't have them
    :param skip_single_underscores: whether to refrain from importing things that are preceded with a single underscore.
        Pertains to modules, as well as items
    :param skip_not_having_all: skip module's not having an __all__ entry
    :raise AttributeError: module's __all__ contained entry that was not in this module
    """
    for importer, modname, ispkg in pkgutil.walk_packages(path, onerror=lambda x: None):
        if recursive and ispkg:
            if modname.startswith('_') and skip_single_underscores:
                continue
            module = importlib.import_module(package_prefix + '.' + modname)
            try:
                mod_all = module.__all__
            except AttributeError:
                if skip_not_having_all:
                    continue
                mod_all = []
                if create_all:
                    module.__all__ = mod_all
            import_from([os.path.join(path[0], modname)], package_prefix + '.' + modname, mod_all,
                        module.__dict__, recursive=recursive,
                        fail_on_attributerror=fail_on_attributerror, create_all=create_all,
                        skip_not_having_all=skip_not_having_all,
                        skip_single_underscores=skip_single_underscores),
            locals[modname] = module
            if modname not in all_:
                all_.append(modname)
        elif not ispkg:
            module = importlib.import_module(package_prefix + '.' + modname)
            try:
                package_ref = module.__all__
            except AttributeError:
                warnings.warn('Module %s does not contain __all__, enumerating it instead' %
                              (package_prefix + '.' + modname, ), RuntimeWarning)
                package_ref = dir(module)

            for item in package_ref:
                if item.startswith('_') and skip_single_underscores:
                    continue
                try:
                    locals[item] = getattr(module, item)
                except AttributeError:
                    if fail_on_attributerror:
                        raise
                else:
                    if item not in all_:
                        all_.append(item)



import subprocess
import typing as tp
import threading

from .exceptions import ProcessFailed

__all__ = ['call_and_return_stdout']


def read_nowait(process: subprocess.Popen, output_list: tp.List[str]) -> None:
    """
    To be launched as a daemon thread. This reads stdout and appends it's entries to a list
    """
    try:
        while True:
            try:
                process.wait(timeout=0.1)
            except subprocess.TimeoutExpired:
                pass
            else:
                line = process.stdout.read(2048)
                if line:
                    output_list.append(line)
                break
            line = process.stdout.read(2048)
            if line == '':
                break
            output_list.append(line)
    except (IOError, OSError):
        pass


def call_and_return_stdout(args: tp.Union[str, tp.List[str]],
                           timeout: tp.Optional[int] = None,
                           encoding: tp.Optional[str] = None,
                           expected_return_code: tp.Optional[int] = None,
                           **kwargs) -> tp.Union[bytes, str]:
    """
    Call a process and return it's stdout.

    Everything in kwargs will be passed to subprocess.Popen

    A bytes object will be returned if encoding is not defined, else stdout will be decoded
    according to specified encoding.

    :param args: arguments to run the program with. Can be either a string or a list of strings.
    :param timeout: amount of seconds to wait for the process result. If process does not complete
        within this time, it will be sent a SIGKILL
    :param encoding: encoding with which to decode stdout. If none is passed, it will be returned as a bytes object
    :param expected_return_code: an expected return code of this process. 0 is the default. If process
        returns anything else, ProcessFailed will be raise. If left default (None) return code won't be checked
        at all
    :raises ProcessFailed: process' result code was different from the requested
    """
    kwargs['stdout'] = subprocess.PIPE

    stdout_list = []

    proc = subprocess.Popen(args, **kwargs)
    reader_thread = threading.Thread(target=read_nowait, args=(proc, stdout_list), daemon=True)
    reader_thread.start()

    try:
        proc.wait(timeout=timeout)
    except subprocess.TimeoutExpired:
        proc.kill()
        proc.wait()
    reader_thread.join()

    if encoding is None:
        result = b''.join(stdout_list)
    else:
        result = ''.join((row.decode(encoding) for row in stdout_list))

    if expected_return_code is not None:
        if proc.returncode != expected_return_code:
            raise ProcessFailed(proc.returncode, result)

    return result




import logging
import typing as tp

T = tp.TypeVar('T')
logger = logging.getLogger(__name__)


_SETTABLE_KEYS = {'_Proxy__obj', '_Proxy__wrap_operations'}


class Proxy(tp.Generic[T]):
    """
    A base class for classes that try to emulate some other object.

    They will intercept all calls and place them on the target object.

    Note that in-place operations will return the Proxy itself, whereas simple addition will shed
    this proxy, returning object wrapped plus something.

    Note that proxies are considered to be the type of the object that they wrap,
    as well as considered to be of type Proxy.

    :param object_to_wrap: object to wrap
    :param wrap_operations: whether results of operations returning something else should be
        also proxied. This will be done by the following code:
        >>> a = a.__add__(b)
        >>> return self.__class__(a)
        Wrapped operations include ONLY add, sub, mul, all kinds of div.
        If you want logical operations wrapped, file an issue.
    """
    __slots__ = ('__obj', '__wrap_operations')


    def __init__(self, object_to_wrap: T, wrap_operations: bool = False):
        self.__obj = object_to_wrap   # type: T
        self.__wrap_operations = wrap_operations

    def __call__(self, *args, **kwargs):
        return self.__obj(*args, **kwargs)

    def __getitem__(self, item):
        return self.__obj[item]

    def __setitem__(self, key, value):
        self.__obj[key] = value

    def __delitem__(self, key):
        del self.__obj[key]

    def __setattr__(self, key, value):
        if key in _SETTABLE_KEYS:
            super().__setattr__(key, value)
        else:
            setattr(self.__obj, key, value)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __delattr__(self, item):
        delattr(self.__obj, item)

    def __int__(self):
        return int(self.__obj)

    def __float__(self):
        return float(self.__obj)

    def __complex__(self):
        return complex(self.__obj)

    def __str__(self):
        return str(self.__obj)

    def __add__(self, other):
        result = self.__obj + other
        if self.__wrap_operations:
            result = self.__class__(result)
        return result

    def __iadd__(self, other):
        self.__obj += other
        return self

    def __sub__(self, other):
        result = self.__obj - other
        if self.__wrap_operations:
            result = self.__class__(result)
        return result

    def __isub__(self, other):
        self.__obj -= other
        return self

    def __mul__(self, other):
        result = self.__obj * other
        if self.__wrap_operations:
            result = self.__class__(result)
        return result

    def __divmod__(self, other):
        result = divmod(self.__obj, other)
        if self.__wrap_operations:
            result = self.__class__(result)
        return result

    def __floordiv__(self, other):
        result = self.__obj // other
        if self.__wrap_operations:
            result = self.__class__(result)
        return result

    def __truediv__(self, other):
        result = self.__obj / other
        if self.__wrap_operations:
            result = self.__class__(result)
        return result

    def __imul__(self, other):
        self.__obj *= other
        return self

    def __itruediv__(self, other):
        self.obj /= other
        return self

    def __ifloordiv__(self, other):
        self.obj //= other
        return self

    def __ilshift__(self, other):
        self.obj <<= other
        return self

    def __irshift__(self, other):
        self.obj >>= other
        return self

    def __iter__(self):
        return iter(self.__obj)

    def __len__(self):
        return len(self.__obj)

    def __contains__(self, item):
        return item in self.__obj

    def __hash__(self):
        return hash(self.__obj)

    def __eq__(self, other):
        return self.__obj == other

    def __or__(self, other):
        return self.__obj or other

    def __and__(self, other):
        return self.__obj and other

    def __le__(self, other):
        return self.__obj <= other

    def __lt__(self, other):
        return self.__obj < other

    def __ge__(self, other):
        return self.__obj >= other

    def __gt__(self, other):
        return self.__obj > other

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.__obj.__exit__(exc_type, exc_val, exc_tb)

    def __repr__(self):
        return repr(self.__obj)

    def __abs__(self):
        return abs(self.__obj)

    def __bool__(self):
        return bool(self.__obj)

    def __format__(self, format_spec):
        return self.__obj.__format__(format_spec)

    def __next__(self):
        return next(self.__obj)

    def __xor__(self, other):
        return self.__obj ^ other



from .registry import register_metric
from .base import Metric
from ..data import MetricDataCollection


@register_metric
class EmptyMetric(Metric):
    """
    A metric that disregards all data that it's fed, and outputs nothing.

    A placeholder for the times when you configure metrics and decide to leave some of them out
    blank.
    """
    __slots__ = ()

    CLASS_NAME = 'empty'

    def _handle(self, *args, **kwargs) -> None:
        pass

    def to_metric_data(self) -> MetricDataCollection:
        return MetricDataCollection()



import typing as tp
import re
import os
__all__ = ['read_re_sub_and_write', 'find_files', 'split']

SEPARATORS = {'\\', '/'}
SEPARATORS.add(os.path.sep)


def _has_separator(path: str) -> bool:
    # handle Windows case
    if len(path) == 3:
        if path.endswith(':/') or path.endswith(':\\'):
            return False
    return any(map(lambda x: x in path, SEPARATORS))


def split(path: str) -> tp.List[str]:
    """
    An exact reverse of os.path.join

    Is is true that

    >>> os.path.join(split(a)) == a
    """
    data = list(os.path.split(path))
    while _has_separator(data[0]):
        data = list(os.path.split(data[0])) + data[1:]
    return data


def read_re_sub_and_write(path: str, pattern: tp.Union[re.compile, str],
                          repl: tp.Union[tp.Callable[[tp.Any], str]]) -> None:
    """
    Read a text file, treat with re.sub and write the contents

    :param path: path of file to treat
    :param pattern: regexp compiled pattern or a string, a pattern to match the file contents
    :param repl: string or a callable(re.Match)->str to replace the contents
    """
    with open(path, 'r') as f_in:
        data = f_in.read()

    if isinstance(pattern, str):
        data = re.sub(pattern, repl, data)
    else:
        data = pattern.sub(repl, data)

    with open(path, 'w') as f_out:
        f_out.write(data)


def _cond_join(prefix: tp.Optional[str], filename: str) -> str:
    """or a conditional os.path.join"""
    if prefix is None:
        return filename
    else:
        return os.path.join(prefix, filename)


def find_files(path: str, wildcard: str = r'(.*)',
               prefix_with: tp.Optional[str] = None,
               scan_subdirectories: bool = True,
               apply_wildcard_to_entire_path: bool = False,
               prefix_with_path: bool = True) -> tp.Iterator[str]:
    """
    Look at given path's files and all subdirectories and return an iterator of
    file names (paths included) that conform to given wildcard.

    Note that wildcard is only applied to the file name if apply_wildcard_to_entire_path is False,
    else the wildcard is applied to entire path (including the application of prefix_with!).

    Files will be additionally prefixed with path, but only if prefix_with_path is True

    :param path: path to look into.
    :param wildcard: a regular expression to match
    :param prefix_with: an optional path component to prefix before the filename with os.path.join
    :param scan_subdirectories: whether to scan subdirectories
    :param apply_wildcard_to_entire_path: whether to take the entire relative path into account
        when checking wildcard
    :param prefix_with_path: whether to add path to the resulting path
    :return: paths with the files. They will be relative paths, relative to path
    """
    if prefix_with_path:
        prefix_with = _cond_join(prefix_with, path)

    for filename in os.listdir(path):
        if scan_subdirectories and os.path.isdir(os.path.join(path, filename)):
            new_prefix = _cond_join(prefix_with, filename)
            yield from find_files(os.path.join(path, filename), wildcard, prefix_with=new_prefix,
                                  prefix_with_path=False)
        else:
            if apply_wildcard_to_entire_path:
                fn_path = _cond_join(prefix_with, filename)
            else:
                fn_path = filename
            if re.match(wildcard, fn_path):
                yield _cond_join(prefix_with, filename)



import typing as tp

from .metric_types import CallableMetric
from .metric_types.measurable_mixin import MeasurableMixin
from .metric_types.base import Metric, MetricLevel


class LabeledMetric(Metric, MeasurableMixin):
    """
    A wrapper to another metric that will always call it's .runtime and .handle with some predefined labels

    Use like:

    >>> a = getMetric('a', 'counter')
    >>> b = LabeledMetric(a, key=5)

    Then this:

    >>> a.runtime(1, key=5)

    Will be equivalent to this:

    >>> b.runtime(1)
    """
    __slots__ = ('metric_to_wrap', 'labels', 'callable')

    def __init__(self, metric_to_wrap, **labels):
        self.metric_to_wrap = metric_to_wrap
        self.labels = labels
        self.callable = labels.pop('callable', lambda: 0)

        if isinstance(metric_to_wrap, CallableMetric):
            metric_to_wrap.register_labeled_metric(self)

    def handle(self, level: tp.Union[int, MetricLevel], *args, **kwargs) -> None:
        kwargs.update(self.labels)
        self.metric_to_wrap.handle(level, *args, **kwargs)





import typing as tp
from collections import namedtuple


def _nth(it, n):
    return [x[n] for x in it]


def _adjust(q):
    var, type_ = q
    if type_ in (None, 'self') or isinstance(var, type_):
        return var
    return type_(var)


def typednamedtuple(cls_name: str, *arg_name_type: type) -> tp.Type[tp.Tuple]:
    """
    Returns a new subclass of tuple with named fields.
    Fields will be coerced to type passed in the pair.

    Parameters are tuples of (field name, class/constructor as callable/1)

    For example:

    >>> tnt = typednamedtuple('tnt', ('x', float), ('y', float))
    >>> a = tnt('5.0', y=2)

    a.x is float, a.y is float too
    """

    fieldnames = []
    typeops = []
    mapping = {}

    for name, type_ in arg_name_type:
        fieldnames.append(name)
        typeops.append(type_)
        mapping[name] = type_

    MyCls = namedtuple(cls_name, fieldnames)

    class Wrapper(MyCls):
        __doc__ = MyCls.__doc__
        __name__ = MyCls.__name__

        def __new__(cls, *args, **kwargs):
            nargs = list(map(_adjust, zip(args, typeops[:len(args)])))

            for next_field_name in fieldnames[len(nargs):]:
                try:
                    nargs.append(_adjust((kwargs.pop(next_field_name), mapping[next_field_name])))
                except KeyError:
                    raise TypeError('Field %s not given', next_field_name)

            if len(kwargs) > 0:
                raise TypeError('Too many parameters')

            return MyCls.__new__(MyCls, *nargs)

    return Wrapper



import threading
import typing as tp

from ..structures.proxy import Proxy
T = tp.TypeVar('T')


class LockedStructure(Proxy, tp.Generic[T]):
    """
    A wizard to make every Python structure thread-safe.

    It wraps obj_to_wrap, passing on all calls, settings and so on to the object wrapper,
    from lock exposing only the context manager protocol.

    Example:

    >>> locked_dict = LockedStructure(dict)
    >>> with locked_dict:
    >>>     locked_dict[5] = 2
    
    Also, please note that operations such as addition will strip this object of being a locked
    structure, ie. they will return object that participated in locked structure plus some other.
    
    Note that in-place operations return the locked structure.
    """
    __slots__ = ('__lock',)

    def __init__(self, obj_to_wrap: T, lock: tp.Optional[threading.Lock] = None):
        super().__init__(obj_to_wrap)
        self.__lock = lock or threading.Lock()

    def __setattr__(self, key, value):
        if key == '_LockedStructure__lock':
            object.__setattr__(self, key, value)
        else:
            super().__setattr__(key, value)

    def __enter__(self):
        self.__lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__lock.release()
        return False



import typing as tp

__all__ = ['merge_dicts']


def merge_dicts(v1: tp.Any, v2: tp.Any) -> tp.Any:
    """
    Try to merge two dicts/list together. If key collision is found, value from v2 will be taken.

    If the objects aren't dicts or lists, v2 will be returned.

    Lists will be concatenated, and dicts updated. v1 will be updated in-place!
    """
    if isinstance(v1, dict) and isinstance(v2, dict):
        for k in v2.keys():
            try:
                v1[k] = merge_dicts(v1[k], v2[k])
            except KeyError:
                v1[k] = v2[k]
        return v1

    if isinstance(v1, list) and isinstance(v2, list):
        v1.extend(v2)
        return v1

    return v2



from .memthread import MemoryPressureManager
import gc


def install_force_gc_collect(severity_level: int = 1) -> None:
    """
    Install a default first severity level handler that forces a GC collection

    :param severity_level: severity level on which to call
    """
    MemoryPressureManager().register_on_entered_severity(severity_level)(gc.collect)



import inspect

import typing as tp
import time

from concurrent.futures import Future

from satella.coding import wraps
from .base import MetricLevel


class MeasurableMixin:
    """
    Add a .measure() method, useful for HistogramMetric and SummaryMetric
    """
    __slots__ = ()

    def measure_future(self, future: Future, logging_level: MetricLevel = MetricLevel.RUNTIME,
                       value_getter: tp.Callable[[], float] = time.monotonic, **labels):
        """
        A function to measure a difference between some value after the method call
        and before it.

        The value will be taken at the moment this function executes, and the moment the future completes
        (with or without an exception)

        :param future: future that is considered
        :param logging_level: one of RUNTIME or DEBUG
        :param value_getter: a callable that takes no arguments and returns a float, which is
            the value
        :param labels: extra labels to call handle() with
        """
        future.old_value = value_getter()

        def on_future_done(fut: Future):
            self.handle(logging_level, value_getter() - fut.old_value, **labels)

        future.add_done_callback(on_future_done)

    def measure(self, include_exceptions: bool = True,
                logging_level: MetricLevel = MetricLevel.RUNTIME,
                value_getter: tp.Callable[[], float] = time.monotonic, **labels):
        """
        A decorator to measure a difference between some value after the method call
        and before it.

        By default, it will measure the execution time.

        Use like:

        >>> call_time = getMetric('root.metric_name.execution_time', 'summary')
        >>> @call_time.measure()
        >>> def measure_my_execution(args):
        >>>     ...

        If wrapped around generator, it will time it from the first element to the last,
        so beware that it will depend on the speed of the consumer.

        It also can be used as a context manager:

        >>> with call_time.measure(logging_level=MetricLevel.DEBUG, label='key'):
        >>>     ...

        :param include_exceptions: whether to include exceptions
        :param logging_level: one of RUNTIME or DEBUG
        :param value_getter: a callable that takes no arguments and returns a float, which is
            the value
        :param labels: extra labels to call handle() with
        """

        class MeasurableMixinInternal:
            def __init__(self, metric_class, include_exceptions, value_getter, logging_level, labels):
                self.metric_class = metric_class
                self.value_getter = value_getter
                self.logging_level = logging_level
                self.include_exceptions = include_exceptions
                self.labels = labels
                self.value = None

            def __call__(self, fun):
                @wraps(fun)
                def inner_normal(*args, **kwargs):
                    start_value = value_getter()
                    excepted = None
                    try:
                        return fun(*args, **kwargs)
                    except Exception as e:
                        excepted = e
                    finally:
                        value_taken = self.value_getter() - start_value
                        if excepted is not None and not self.include_exceptions:
                            raise excepted

                        self.metric_class.handle(logging_level, value_taken, **labels)

                        if excepted is not None:
                            raise excepted

                @wraps(fun)
                def inner_generator(*args, **kwargs):
                    start_value = value_getter()
                    excepted = None
                    try:
                        yield from fun(*args, **kwargs)
                    except Exception as e:
                        excepted = e
                    finally:
                        value_taken = value_getter() - start_value
                        if excepted is not None and not self.include_exceptions:
                            raise excepted

                        self.metric_class.handle(self.logging_level, value_taken, **self.labels)

                        if excepted is not None:
                            raise excepted

                if inspect.isgeneratorfunction(fun):
                    return inner_generator
                else:
                    return inner_normal

            def __enter__(self):
                self.value = self.value_getter()
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                if not self.include_exceptions and exc_type is not None:
                    return False

                elapsed = self.value_getter() - self.value
                self.metric_class.handle(self.logging_level, elapsed, **self.labels)

                return False

        return MeasurableMixinInternal(self, include_exceptions, value_getter, logging_level, labels)



import os
import sys
import typing as tp
import types

try:
    import pwd
    import grp
except ImportError:
    # Windows?
    class L(object):  # always return UID=GID=0
        __slots__ = ()

        def __getattr__(self, item):
            return lambda q: 0

    pwd = grp = L()

DEVNULL = '/dev/null'

__all__ = ['daemonize']


def daemonize(exit_via: tp.Callable = sys.exit,
              redirect_std_to_devnull: bool = True,
              uid: tp.Optional[int] = None,
              gid: tp.Optional[int] = None):
    """
    Make this process into a daemon.

    This entails:

    - umask 0
    - forks twice
    - be the child of init
    - becomes session leader
    - changes root directory to /
    - closes stdin, stdout, stderr
    - (option) redirects stdin, stdout, stderr to /dev/null

    Refer - "Advanced Programming in the UNIX Environment" 13.3

    :param exit_via: callable used to terminate process
    :param redirect_std_to_devnull: whether to redirect stdin, stdout and
        stderr to /dev/null
    :param uid: User to set (via seteuid). Default - this won't be done. You
        can pass either user name as string or UID.
    :param gid: Same as UID, but for groups. These will be resolved too.
    :raises KeyError: uid/gid was passed as string, but getpwnam() failed
    :raises OSError: platform is Windows
    """

    if sys.platform.startswith('win'):
        raise OSError('Cannot call daemonize on Windows!')

    _double_fork(exit_via=exit_via)
    _close_descriptors()

    if redirect_std_to_devnull:
        _redirect_descriptors_to_null()

    _parse_ug(uid, pwd, 'pw_uid', os.seteuid)
    _parse_ug(gid, grp, 'gr_gid', os.setegid)


def _parse_ug(no: tp.Union[str, int], module: types.ModuleType, field_name: str,
              osfun: tp.Callable[[int], None]) -> None:
    if no is not None:
        if isinstance(no, str):
            no = getattr(module.getpwnam(no), field_name)
        osfun(no)


def _redirect_descriptors_to_null() -> None:
    sys.stdin = open(DEVNULL, 'rb')
    sys.stdout = open(DEVNULL, 'wb')
    sys.stderr = open(DEVNULL, 'wb')


def _close_descriptors() -> None:
    for d in [sys.stdin, sys.stdout, sys.stderr]:
        d.close()


def _double_fork(exit_via: tp.Callable[[], None]) -> None:
    os.umask(0)

    if os.fork() > 0:
        exit_via()  # parent exits

    os.setsid()

    if os.fork() > 0:
        exit_via()  # parent exits
    os.chdir('/')



import copy
import re
import typing as tp

from satella.coding.concurrent.callablegroup import CallableGroup
from satella.coding.recast_exceptions import rethrow_as
from satella.imports import import_class
from ..exceptions import ConfigurationValidationError, ConfigurationSchemaError

__all__ = [
    'Descriptor',
    'Integer', 'Float', 'String', 'Boolean',
    'IPv4',
    'Regexp',
    'List', 'Dict', 'Union',
    'create_key',
    'must_be_type',
    'must_be_one_of',
    'Caster',
    'CheckerCondition',
    'ConfigDictValue',
    'descriptor_from_dict',
    'register_custom_descriptor',
]

ConfigDictValue = tp.Optional[tp.Union[int, float, str, dict, list, bool]]
CheckerConditionType = tp.Callable[[ConfigDictValue], bool]
ObjectMakerType = tp.Callable[
    [ConfigDictValue], tp.Any]  # might raise ConfigurationSchemaError as well


class CheckerCondition:
    __slots__ = ('condition', 'description', 'is_pre_checker')

    PRE_CHECKER = 0
    POST_CHECKER = 1

    def __init__(self, condition: CheckerConditionType, description: str = u'',
                 is_pre_checker: bool = True):
        self.condition = condition
        self.description = description
        self.is_pre_checker = is_pre_checker

    def __call__(self, value):
        if not self.condition(value):
            raise ConfigurationValidationError(self.description, value)


def must_be_type(*cls_or_tuple):
    return CheckerCondition(condition=lambda v: isinstance(v, cls_or_tuple),
                            description='not one of types %s' % (cls_or_tuple,))


def must_be_one_of(*items):
    return CheckerCondition(condition=lambda v: v in items,
                            description='not in set %s' % (items,),
                            is_pre_checker=False)


class Descriptor:
    """
    Base class for a descriptor
    """
    __slots__ = ('pre_checkers', 'post_checkers', 'name', 'optional', 'default',
                 'my_exceptions')

    BASIC_MAKER = staticmethod(lambda v: v)
    MY_EXCEPTIONS = [TypeError, ValueError]  # a list of Exception classes
    CHECKERS = []  # a list of CheckerCondition

    def __init__(self):
        self.pre_checkers = CallableGroup()     # type: tp.Callable[[bool], None]
        self.post_checkers = CallableGroup()    # type: tp.Callable[[bool], None]
        self.name = None                        # type: tp.Optional[str]
        self.optional = None                    # type: tp.Optional[bool]
        self.default = None                     # type: tp.Optional[tp.Any]

        for checker in self.__class__.CHECKERS:
            self.add_checker(checker)

        self.my_exceptions = tuple(self.MY_EXCEPTIONS)  # type: tp.Tuple[tp.Type[Exception], ...]

    def __str__(self):
        return '%s()' % (self.__class__.__qualname__, )

    def __call__(self, value: ConfigDictValue) -> tp.Any:
        """
        raises ConfigurationSchemaError: on invalid schema
        """
        self.pre_checkers(value)

        try:
            value = self.BASIC_MAKER(value)
        except self.my_exceptions as e:
            raise ConfigurationValidationError('could not pass to maker due to %s' % (e, ), value)

        self.post_checkers(value)

        return value

    def add_checker(self, checker: CheckerCondition):
        if checker.is_pre_checker:
            self.pre_checkers.add(checker)
        else:
            self.post_checkers.add(checker)


# noinspection PyDecorator
@staticmethod
def _make_boolean(v: tp.Any) -> bool:
    if isinstance(v, str):
        if v.upper() == 'TRUE':
            return True
        elif v.upper() == 'FALSE':
            return False
        else:
            raise ConfigurationValidationError('Unknown value of "%s" posing to be a bool'
                                               % (v,))
    else:
        return bool(v)


class Boolean(Descriptor):
    """
    This value must be a boolean, or be converted to one
    """
    BASIC_MAKER = _make_boolean


class Integer(Descriptor):
    """
    This value must be an integer, or be converted to one
    """
    BASIC_MAKER = int


class Float(Descriptor):
    """
    This value must be a float, or be converted to one
    """
    BASIC_MAKER = float


class String(Descriptor):
    """
    This value must be a string, or be converted to one
    """
    BASIC_MAKER = str


class Regexp(String):
    """
    Base class for declaring regexp-based descriptors. Overload it's attribute REGEXP. Use as
    following:

    >>> class IPv6(Regexp):
    >>>     REGEXP = '(([0-9a-f]{1,4}:)' ...
    """
    __slots__ = ('regexp', )

    REGEXP = r'.*'

    def __init__(self):
        super().__init__()
        if isinstance(self.REGEXP, str):
            self.regexp = re.compile(self.REGEXP)
        else:
            self.regexp = self.REGEXP

    def __call__(self, value: ConfigDictValue) -> str:
        value = super(Regexp, self).__call__(value)

        match = self.regexp.match(value)
        if not match:
            raise ConfigurationValidationError('value does not match %s' % (self.REGEXP.pattern,),
                                               value)

        return match.group(0)

    def __str__(self):
        return 'Regexp(%s)' % (self.REGEXP.pattern, )


class IPv4(Regexp):
    """
    This must be a valid IPv4 address (no hostnames allowed)
    """
    REGEXP = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'


class List(Descriptor):
    """
    This must be a list, made of entries of a descriptor (this is optional)
    """
    __slots__ = ('type_descriptor', )

    CHECKERS = [must_be_type(list, tuple)]
    BASIC_MAKER = list

    def __init__(self, type_descriptor: tp.Optional[Descriptor] = None):
        super().__init__()
        self.type_descriptor = type_descriptor or Descriptor()      # type: Descriptor

    def __call__(self, value: ConfigDictValue) -> tp.List:
        value = super().__call__(value)

        return [self.type_descriptor(p) for p in value]

    def __str__(self):
        return 'List(%s)' % (self.type_descriptor, )


DictDescriptorKey = tp.NewType('DictDescriptorKey', Descriptor)


def create_key(descriptor: Descriptor, name: str, optional: bool = False,
               default: tp.Optional[tp.Any] = None) -> DictDescriptorKey:
    descriptor.name = name
    descriptor.optional = optional
    descriptor.default = default

    return descriptor


class Caster(Descriptor):
    """
    A value must be ran through a function.

    Use like:

    >>> class Environment(enum.IntEnum):
    >>>     PRODUCTION = 0
    >>> assert Caster(Environment)(0) == Environment.PRODUCTION
    """

    def __init__(self, to_cast: tp.Callable[[tp.Any], tp.Any]):
        self.to_cast = to_cast

    @rethrow_as(ValueError, ConfigurationValidationError)
    def __call__(self, value: ConfigDictValue):
        return self.to_cast(value)


class Dict(Descriptor):
    """
    This entry must be a dict, having at least specified keys.

    Use like:

    >>> Dict([
    >>>     create_key(String(), 'key_s'),
    >>>     create_key(Integer(), 'key_i'),
    >>>     create_key(Float(), 'key_f'),
    >>>     create_key(String(), 'key_not_present', optional=True,
    >>>                default='hello world'),
    >>>     create_key(IPv4(), 'ip_addr')
    >>>])
    """
    __slots__ = ('keys', 'unknown_key_mapper')

    BASIC_MAKER = dict
    CHECKERS = [must_be_type(dict)]

    UnknownKeyHandlerType = tp.Callable[[str, ConfigDictValue], tp.Any]

    def __init__(self, keys: tp.List[DictDescriptorKey],
                 unknown_key_mapper: UnknownKeyHandlerType = lambda str,
                                                                    data: data):
        super().__init__()
        self.keys = {item.name: item for item in keys}  # type: tp.Dict[str, DictDescriptorKey]
        self.unknown_key_mapper = unknown_key_mapper  # Dict.UnknownKeyHandlerType

    def __str__(self):
        return 'Dict(%s)' % (self.keys, )

    def __call__(self, value: ConfigDictValue) -> dict:
        value = copy.copy(value)
        if not isinstance(value, dict):
            raise ConfigurationValidationError('value passed was not a dict: %s' % (value,))
        value = super().__call__(value)

        if not isinstance(value, dict):
            raise ConfigurationValidationError('value received from descriptor was not a dict: %s'
                                               % (value,))

        output = {}

        for key, key_descriptor in self.keys.items():
            try:
                v = value.pop(key)
            except KeyError:
                if key_descriptor.optional:
                    output[key] = key_descriptor.default
                else:
                    raise ConfigurationValidationError('required key %s not found' % (key,))
            else:
                output[key] = key_descriptor(v)

        for k, v in value.items():
            output[k] = self.unknown_key_mapper(k, v)

        return output


class Union(Descriptor):
    """
    The type of one of the child descriptors. If posed as such:

    Union(List(), Dict())

    then value can be either a list or a dict
    """
    __slots__ = ('descriptors', )

    def __init__(self, *descriptors: tp.List[Descriptor]):
        super().__init__()
        self.descriptors = descriptors

    def __call__(self, value: tp.Any) -> tp.Any:
        for descriptor in self.descriptors:
            try:
                return descriptor(value)
            except ConfigurationValidationError:
                continue
        raise ConfigurationValidationError('Could not match value %s to a descriptor' % (value,))


BASE_LOOKUP_TABLE = {'int': Integer, 'float': Float, 'str': String, 'ipv4': IPv4, 'list': List,
                     'dict': Dict, 'any': Descriptor, 'bool': Boolean, 'union': Union,
                     'caster': Caster}


def _get_descriptor_for(key: str, value: tp.Any) -> Descriptor:
    if value == '':
        return Descriptor()
    elif isinstance(value, str):
        if value in ('int', 'float', 'str', 'ipv4', 'any', 'bool'):
            return create_key(BASE_LOOKUP_TABLE[value](),
                              key, False, None)
    elif isinstance(value, dict):
        if 'type' not in value:
            return create_key(descriptor_from_dict(value),
                              key, False, None)
        else:
            args = ()
            type_ = value['type']
            if type_ == 'list':
                of = _get_descriptor_for('', value.get('of', ''))
                args = (of,)
            elif type_ == 'caster':
                cast_to = value.get('cast_to')
                dict_to_look_in = globals().copy()
                if not isinstance(__builtins__, dict):
                    dict_to_look_in.update(__builtins__.__dict__)
                else:
                    dict_to_look_in.update(__builtins__)
                dict_to_look_in.update(locals())
                if cast_to in dict_to_look_in:
                    y = dict_to_look_in[cast_to]
                else:
                    y = import_class(cast_to)

                if 'expr' in value:
                    dict_ = globals().copy()
                    dict_.update(locals())
                    dict_['y'] = y
                    args = eval('lambda x: ' + value['expr'], dict_),
                else:
                    args = y,
            elif type_ == 'union':
                args = [_get_descriptor_for('', x) for x in value.get('of', [])]
            optional, default = False, None
            if 'default' in value:
                optional = True
                default = value['default']
            optional = value.get('optional', optional)
            descriptor = BASE_LOOKUP_TABLE[type_](*args)
            return create_key(descriptor, key, optional, default)
    elif isinstance(value, type):
        return _get_descriptor_for(key, value.__qualname__)
    else:
        raise ConfigurationSchemaError('invalid schema, unrecognized config object %s' % (value,))


def register_custom_descriptor(name: str):
    """
    A decorator used for registering custom descriptors in order to be loadable via
    descriptor_from_dict

    Use like:

    >>> @register_custom_descriptor('ipv6')
    >>> class IPv6(Regexp):
    >>>     REGEXP = '(([0-9a-f]{1,4}:)' ...

    :param name: under which it is supposed to be invokable
    """

    def inner(cls):
        BASE_LOOKUP_TABLE[name] = cls
        return cls

    return inner


def descriptor_from_dict(dct: dict) -> Descriptor:
    """
    Giving a Python dictionary-defined schema of the configuration, return a Descriptor-based one

    :param dct: something like
    {
        "a": "int",
        "b": "str",
        "c": {
            "type": "int"
            "optional": True,
            "default": 5
        },
        "d": {
            "a": "int",
            "b": "str"
        }
    }

    although you can pass "int", "float" and "str" without enclosing quotes, that will work too

    :return: a Descriptor-based schema
    """
    fields = []

    for key, value in dct.items():
        descriptor = _get_descriptor_for(key, value)
        fields.append(descriptor)

    return Dict(fields)
